/*
  ddless
  
  Instead of dd if=/dev/source of=/dev/target, we do this once, while reading we
  record segment (16k) signatures. Next time, we read, compare the signature, if
  nothing change, we continue to read. Otherwise, we update the signature and
  write to the target.
  
  There is a reader and ddless thread which decouples the disk read from the
  signature calculation. With single threaded code we achieved 100MB/s read
  throughput, while the threaded version read at 150MB/s.
  
  Steffen Plotner, 2008
*/

#define _LARGEFILE64_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <asm/fcntl.h>
#include <sys/mman.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <utime.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include <zlib.h>

#include "dd_log.h"
#include "dd_murmurhash2.h"
#include "dd_file.h"

#define RELEASE_DATE "2008.08.03"

#define READ_BUFFERS 32
#define READ_BUFFER_SIZE (8 * 1024 * 1024)
#define SEGMENT_SIZE (16*1024)
#define BUFFER_SEGMENTS READ_BUFFER_SIZE/SEGMENT_SIZE

#define GIGABYTE_FACTOR (1024 * 1024 * 1024)
#define MEGABYTE_FACTOR (1024 * 1024)

#define RUNMODE_SOURCE_TARGET 0
#define RUNMODE_CHECKSUM_ONLY 1
#define RUNMODE_NEW_CHECKSUM 2
#define RUNMODE_DDZONE 3

//
// Buffer structure, there are n buffers that are read by one thread,
// and then via a signal, passed onto the ddless thread to compute the
// checksum and perform any necessary writes.
//
#define BUFFER_STATE_READING 0
#define BUFFER_STATE_DDLESS 1
#define BUFFER_STATE_FREE 2
#define BUFFER_STATE_DONE 3
struct buffer_struct
{
	//
	// a buffer mutex and read/ddless thread conditional variables
	//
	pthread_mutex_t	buffer_mutex;
	pthread_cond_t	read_cv;
	pthread_cond_t	ddless_cv;
	int				state;

	//
	// work buffers for reading the source dev
	//
	void		*aligned_buffer;
	int			read_size;

	//
	// track which segment within a buffer has changes (for a 1MB buffer and
	// 16K segment we have 64 segments). One segment tagged at the end -> EOS
	//
	int			seg_bytes_dirty_map[BUFFER_SEGMENTS+1];

	//
	// statistics (at the end the stats are summed up across of all READ_BUFFERS)
	//
	u_int64_t	read_blocks;
	u_int64_t	changed_segments;
	u_int64_t	written_bytes;
};
static struct buffer_struct buffers[READ_BUFFERS];

void *reader_thread();
void *ddless_thread();
void dd_zone_reader();

#define DEV_NAME_LENGTH 1024
#define MAX_CMD_LENGTH 1024

//
// instead of global variables we use a structure
//
struct parms_struct
{
	int			runmode;
	int			o_direct;

	time_t		start_time;
	time_t		end_time;
	
	char		source_dev[DEV_NAME_LENGTH];
	int			source_fd;
	off64_t		source_size_bytes;
	u_int64_t	source_segments;
	u_int64_t	source_segment_remainder_bytes;
	int			max_read_mb_sec;

	pthread_t	reader_thread;
	pthread_t	ddless_thread;

	// work buffers for managing the checksum file (memory mapped)
	char		checksum_file[DEV_NAME_LENGTH];
	int			checksum_file_new;
	u_int64_t	mmap_size;
	int			mmap_fd;
	struct checksum_struct *checksum_array;

	char		stats_file[DEV_NAME_LENGTH];
	int			stats_fd;
	
	char		target_dev[DEV_NAME_LENGTH];
	int			target_fd;
	off64_t		target_size_bytes;
	
	// allow us to inject an OS level command when max_write_bytes is reached
	int			max_enabled;
	u_int64_t	max_write_bytes;
	u_int64_t	max_written_bytes;
	char		max_reached_cmd[MAX_CMD_LENGTH];
	int			max_reached;
};
static struct parms_struct parms;

//
// checksum structure (can accomodate multiple, pluggable algorithms)
//
struct checksum_struct
{
	u_int32_t	checksum1_murmur;
	u_int32_t	checksum2_crc32;
};

//-----------------------------------------------------------------------------
// ddless setup and thread spawn
//-----------------------------------------------------------------------------
int ddless(int runmode)
{
	int i;
	parms.runmode = runmode;
	
	//
	// allocate page size aligned work buffers (required by O_DIRECT)
	//
	dd_log(LOG_INFO, "%d read buffers, %u MB/buffer, read buffer total: %u MB",
		READ_BUFFERS, 
		READ_BUFFER_SIZE / MEGABYTE_FACTOR,
		(READ_BUFFERS * READ_BUFFER_SIZE) / MEGABYTE_FACTOR);
	for(i=0; i<READ_BUFFERS; i++)
	{
		if (posix_memalign((void**)&buffers[i].aligned_buffer, getpagesize(), READ_BUFFER_SIZE))
		{
			dd_log(LOG_ERR, "unable to allocate buffers[%d] with READ_BUFFER_SIZE=%d",i,READ_BUFFER_SIZE);
			exit(1);
		}
		dd_log(LOG_DEBUG, "buffers[%d] aligned buffer: %p",i, buffers[i].aligned_buffer);
	}
	dd_log(LOG_INFO, "segment size: %u bytes", SEGMENT_SIZE);
	dd_log(LOG_INFO, "direct IO: %d", parms.o_direct);

	//
	// setup max byte parameter request
	//
	parms.max_enabled = 0;
	if ( parms.max_write_bytes != 0 )
	{
		parms.max_reached = 0;
		parms.max_written_bytes = 0;
		
		dd_log(LOG_INFO, "maximum %llu bytes to write before invoking: %s",
			parms.max_write_bytes,
			parms.max_reached_cmd);
		if ( *parms.max_reached_cmd == '\0' )
		{
			dd_log(LOG_ERR, "since -m was specified, -i must also be specified");
			exit(1);
		}
		parms.max_enabled = 1;
	}

	//
	// open source device readonly
	//
	if ((parms.source_fd = dd_dev_open_ro(parms.source_dev, parms.o_direct)) == -1 )
	{
		dd_log(LOG_ERR, "unable to open source device: %s",parms.source_dev);
		exit(1);
	}
	dd_log(LOG_INFO, "opened source device: %s",parms.source_dev);
	
	//
	// source device size
	//
	if ((parms.source_size_bytes = dd_device_size(parms.source_fd)) == -1 )
	{
		dd_log(LOG_ERR, "unable to determine source device %s size",parms.source_dev);
		exit(1);
	}
	dd_log(LOG_INFO, "source device size: %llu bytes (%0.2f GB)", 
		parms.source_size_bytes,
		(float)parms.source_size_bytes/GIGABYTE_FACTOR);

	//
	// ddzone ?
	//
	if ( runmode == RUNMODE_DDZONE )
	{
		dd_zone_reader();
		return 0;
	}

	//
	// calculate number of segments based on source size/SEGMENT_SIZE
	//
	parms.source_segments = parms.source_size_bytes / SEGMENT_SIZE;
	parms.source_segment_remainder_bytes = parms.source_size_bytes % SEGMENT_SIZE;
	dd_log(LOG_INFO, "source segments: %llu, remainder: %llu bytes",
		parms.source_segments,parms.source_segment_remainder_bytes );
	if ( parms.source_segment_remainder_bytes > 0 )
	{
		parms.source_segments++;
		dd_log(LOG_INFO, "source segments: %llu (adjusted due to remainder)",
			parms.source_segments);
	}

	//
	// read rate controlled?
	//
	if ( parms.max_read_mb_sec > 0 )
	{
		dd_log(LOG_INFO, "source device max read rate: %d MB/s",parms.max_read_mb_sec);
	}

	//
	// forced checksum run mode, remove the checksum and let it get created below
	//
	if ( runmode == RUNMODE_CHECKSUM_ONLY )
	{
		if ( dd_file_exists(parms.checksum_file) )
		{
			if ( unlink(parms.checksum_file) == -1 )
			{
				dd_log(LOG_ERR, "failed to unlink %s",parms.checksum_file);
				exit(1);
			}
		}
	}
	
	//
	// create/open memory mapped file for segment checksums
	//
	parms.mmap_size = sizeof(struct checksum_struct) * parms.source_segments;
	dd_log(LOG_INFO, "checksum file size: %llu bytes", parms.mmap_size);

	parms.checksum_file_new = 0;
	if ( !dd_file_exists(parms.checksum_file) ||
		dd_file_size(parms.checksum_file) != parms.mmap_size)
	{
		if ( runmode == RUNMODE_NEW_CHECKSUM )
		{
			dd_log(LOG_INFO,"new checksum file required, exiting...");
			exit(3);
		}
		dd_log(LOG_INFO,"new checksum file required");

		//
		// we need to know about this state, when we do the initial 
		// read, checksum calc (and cannot compare it with mmap's data)
		//
		parms.checksum_file_new = 1;
	}

	if ( runmode == RUNMODE_NEW_CHECKSUM )
	{
		dd_log(LOG_INFO,"checksum file will be used, exiting");
		return 0;
	}
	
	if ((parms.mmap_fd = open(parms.checksum_file, O_CREAT|O_RDWR|O_LARGEFILE,
			(mode_t)0600)) == -1 )
	{
		dd_log(LOG_ERR, "unable to open checksum file: %s",parms.checksum_file);
		exit(1);
	}

	//
	// set length of file, causing it on create to be a sparse file checksum file
	// and this is required for mmap'd files, else you have a bus error!
	//
	if ( ftruncate64(parms.mmap_fd,parms.mmap_size) )
	{
		dd_log(LOG_ERR, "ftruncate64 of checksum file %s failed to %llu bytes",
			parms.checksum_file,parms.mmap_size);
		exit(1);
	}

	//
	// mmap checksum file
	//
	parms.checksum_array = mmap64(
		0, parms.mmap_size, PROT_READ | PROT_WRITE, MAP_SHARED, parms.mmap_fd, 0);
	if ( parms.checksum_array == MAP_FAILED )
	{
		dd_log(LOG_ERR, "unable to mmap");
		exit(1);
	}
	dd_log(LOG_INFO,"checksum array ptr: %p", parms.checksum_array);

/* sample mmap write code (check for errors)
	u_int32_t i;
	struct checksum_struct *checksum_ptr;
	checksum_ptr = parms.checksum_array;
	for(i=0; i < parms.source_segments; i++)
	{
		checksum_ptr->checksum = 0;
		checksum_ptr++;
	}
	munmap(0, parms.mmap_size);
	close(parms.mmap_fd);
*/
	if ( runmode == RUNMODE_SOURCE_TARGET )
	{
		//
		// support devices and regular files, files are setup sparse, however, the initial
		// copy will make them non-sparse. If the file exists, then it must be of sufficient
		// size - we don't unlink the file, since we don't unlink a device either, we use it
		// and write.
		//
		if ( !dd_file_exists(parms.target_dev) )
		{
			dd_log(LOG_INFO,"target file %s not found, creating it",parms.target_dev);

			int tmp_fd;
			if ((tmp_fd = open(parms.target_dev, O_WRONLY|O_CREAT|O_LARGEFILE, (mode_t)0600)) == -1 )
			{
				dd_log(LOG_ERR, "unable to create target file: %s",parms.target_dev);
				exit(1);
			}
			close(tmp_fd);

			//
			// since the target file did not exist, we must work from now on as
			// if the checksum was new (i.e. all data gets written)
			//
			parms.checksum_file_new = 1;
		}

		//
		// open the target (o_direct is not sensible for writing)
		//
		if ((parms.target_fd = dd_dev_open_rw(parms.target_dev, 0)) == -1 )
		{
			dd_log(LOG_ERR, "unable to open target device: %s",parms.target_dev);
			exit(1);
		}
		dd_log(LOG_INFO, "opened target device: %s",parms.target_dev);

		//
		// target device size
		//
		if ((parms.target_size_bytes = dd_device_size(parms.target_fd)) == -1 )
		{
			dd_log(LOG_ERR, "unable to determine target device %s size",parms.target_dev);
			exit(1);
		}
		dd_log(LOG_INFO, "target device size: %llu bytes (%0.2f GB)", 
			parms.target_size_bytes,
			(float)parms.target_size_bytes/GIGABYTE_FACTOR);


		//
		// only check target against source size when a non-regular file is involved
		//
		struct stat64 fstat;
		if ( stat64(parms.target_dev,&fstat) )
		{
			dd_log(LOG_ERR, "stat of target device %s failed",parms.target_dev);
			exit(1);
		}
		if ( S_ISREG(fstat.st_mode))
		{
			if ( ftruncate64(parms.target_fd,parms.source_size_bytes) )
			{
				dd_log(LOG_ERR, "ftruncate64 of target device %s failed to %llu bytes",
					parms.target_dev,parms.source_size_bytes);
				exit(1);
			}
		}
		else
		{
			//
			// target must be at least as large as source
			//
			if ( parms.target_size_bytes < parms.source_size_bytes )
			{
				dd_log(LOG_ERR, "target size insufficient");
				exit(1);
			}
		}
	}

	//
	// configure mutexes and conditional variables
	// spawn a reader and ddless processing thread
	//
	for(i=0; i<READ_BUFFERS; i++)
	{
		pthread_mutex_init(&buffers[i].buffer_mutex,NULL);
		pthread_cond_init(&buffers[i].read_cv,NULL);
		pthread_cond_init(&buffers[i].ddless_cv,NULL);
		buffers[i].state = BUFFER_STATE_FREE;
	}

	pthread_attr_t thread_attributes;
	
	pthread_attr_init(&thread_attributes);
	pthread_attr_setdetachstate(&thread_attributes, PTHREAD_CREATE_JOINABLE);

	time(&parms.start_time);

	// threads are running now
	dd_log(LOG_INFO,"launching reader and ddless threads...");
	pthread_create(&parms.ddless_thread, &thread_attributes, ddless_thread, NULL);
	pthread_create(&parms.reader_thread, &thread_attributes, reader_thread, NULL);

	// wait for them to finish
	pthread_join(parms.reader_thread, NULL);
	pthread_join(parms.ddless_thread, NULL);

	time(&parms.end_time);

	pthread_attr_destroy(&thread_attributes);
	for(i=0; i<READ_BUFFERS; i++)
	{
		pthread_mutex_destroy(&buffers[i].buffer_mutex);
		pthread_cond_destroy(&buffers[i].read_cv);
		pthread_cond_destroy(&buffers[i].ddless_cv);
	}

	//
	// performance summary
	//
	u_int64_t read_blocks = 0;
	u_int64_t changed_segments = 0;
	u_int64_t written_bytes = 0;
	for(i=0; i<READ_BUFFERS; i++)
	{
		read_blocks += buffers[i].read_blocks;
		changed_segments += buffers[i].changed_segments;
		written_bytes += buffers[i].written_bytes;
	}
	read_blocks--;
	
	double segment_change_percentage = 
		100*((double)changed_segments/(double)parms.source_segments);

	dd_log(LOG_INFO,"total blocks read %llu",read_blocks);
	dd_log(LOG_INFO,"found changed segments %llu (%0.2f%%) of %llu segments",
		changed_segments,
		segment_change_percentage,
		parms.source_segments);
	if ( runmode == RUNMODE_SOURCE_TARGET )
	{
		dd_log(LOG_INFO,"wrote %llu bytes (%0.2f GB) to target",
			written_bytes, (double)(written_bytes) / GIGABYTE_FACTOR);
	}

	double elapsed_sec = difftime(parms.end_time,parms.start_time);
	
	dd_log(LOG_INFO,"processing time: %0.2f seconds (%0.2f minutes)", 
		elapsed_sec, elapsed_sec/60);
	double processing_perf = (double)(parms.source_size_bytes/1024/1024)/elapsed_sec;

	dd_log(LOG_INFO,"processing performance: %0.2f MB/s",processing_perf);

	//
	// output statistics file
	//
	if ( runmode == RUNMODE_SOURCE_TARGET )
	{
		snprintf(parms.stats_file, DEV_NAME_LENGTH, "%s.stats",parms.checksum_file);
		dd_log(LOG_INFO,"preparing stats file %s", parms.stats_file);
		if ((parms.stats_fd = open(parms.stats_file, O_WRONLY|O_CREAT|O_APPEND, (mode_t)0600)) == -1 )
		{
			dd_log(LOG_ERR, "unable to open stats file: %s",parms.stats_file);
			exit(1);
		}
		char tmp[512];
		struct tm *begin;
		begin = localtime(&parms.start_time);
		snprintf(tmp,512, "%04d-%02d-%02d %02d:%02d:%02d %0.5f segment_change_ratio "
			"%llu bytes_written %0.2f seconds %0.2f MB/s\n",
			1900+begin->tm_year,begin->tm_mon+1,begin->tm_mday,
			begin->tm_hour,begin->tm_min,begin->tm_sec,
			segment_change_percentage/100,
			written_bytes,
			elapsed_sec,
			processing_perf);
		write(parms.stats_fd,tmp,strlen(tmp));
		close(parms.stats_fd);
	}
	
	//
	// dump to stderr dd like stats (can only report blocks read, the
	// amount written varies
	//
	fprintf(stderr,"%llu MB processed\n", 
		parms.source_size_bytes / MEGABYTE_FACTOR);
	
	//
	// close source
	//
	close(parms.source_fd);

	//
	// close memory mapped checksum file (must update the file date/time
	// stamp since mmap does not (http://lkml.org/lkml/2007/2/20/255) and
	// backup programs would then miss the checksum file - shit!)
	//
	munmap(parms.checksum_array, parms.mmap_size);
	close(parms.mmap_fd);
	utime(parms.checksum_file,NULL);

	//
	// close target
	//
	if ( runmode == RUNMODE_SOURCE_TARGET )
		close(parms.target_fd);

	return 0;
}

//-----------------------------------------------------------------------------
// read disk zones
//-----------------------------------------------------------------------------
void dd_zone_reader()
{
	struct timeval	buffer_time_start;
	struct timeval	buffer_time_end;
	long elapsed_seconds;
	long elapsed_useconds;
	long elapsed_mtime;

	//
	// enumerate over the buffers
	//
	gettimeofday(&buffer_time_start,NULL);
	int buffer_read = 0;
	int is_done = 0;
	int i = 0;
	while(!is_done)
	{		
		//
		// read data
		//
		dd_log(LOG_DEBUG,"buffers[%d] read",i);
		buffer_read += read(parms.source_fd, 
			buffers[i].aligned_buffer, READ_BUFFER_SIZE);

		if ( buffer_read == -1 )
		{
			dd_log(LOG_ERR, "unable to read from source device");
			exit(1);
		}

		if ( buffer_read == 0 )
		{
			is_done = 1;
		}

		//
		// next buffer
		//
		i++;
		i = i % READ_BUFFERS;

		if ( i == 0 )
		{
			gettimeofday(&buffer_time_end,NULL);

			elapsed_seconds  = buffer_time_end.tv_sec  - buffer_time_start.tv_sec;
			elapsed_useconds = buffer_time_end.tv_usec - buffer_time_start.tv_usec;
			elapsed_mtime = ((elapsed_seconds) * 1000 + elapsed_useconds/1000.0) + 0.5;
			
			printf("%u MB %ld ms %0.2f MB/s\n", 
				buffer_read / MEGABYTE_FACTOR,
				elapsed_mtime,
				(double)((READ_BUFFERS * READ_BUFFER_SIZE)/MEGABYTE_FACTOR) / ((double)elapsed_mtime/1000));
			fflush(stdout);
			gettimeofday(&buffer_time_start,NULL);
			buffer_read = 0;
		}
	}
}

//-----------------------------------------------------------------------------
// reader thread
//-----------------------------------------------------------------------------
void *reader_thread()
{
	struct timeval	buffer_time_start;
	struct timeval	buffer_time_end;
	long elapsed_seconds;
	long elapsed_useconds;
	long elapsed_mtime;
	double read_mb_sec;
	int sleep_factor_usec = 0;
	
	//
	// enumerate over the buffers
	//
	gettimeofday(&buffer_time_start,NULL);
	int is_done = 0;
	int i = 0;
	while(!is_done)
	{

		//
		// lock buffer
		//
		pthread_mutex_lock(&buffers[i].buffer_mutex);
		dd_log(LOG_DEBUG,"th_reader buffers[%d] lock",i);
		
		//
		// we have a lock and if we find that it is NOT free, then we need
		// to wait (the wait does an implicit unlock, so that the other
		// thread can do a lock, process, signal)
		//
		while ( buffers[i].state != BUFFER_STATE_FREE )
		{
			dd_log(LOG_DEBUG,"th_reader buffers[%d] cond_wait",i);
			pthread_cond_wait(&buffers[i].read_cv, &buffers[i].buffer_mutex);
		}

		//
		// read data
		//
		dd_log(LOG_DEBUG,"th_reader buffers[%d] read",i);
		buffers[i].read_size = read(parms.source_fd, 
			buffers[i].aligned_buffer, READ_BUFFER_SIZE);

		if ( buffers[i].read_size == -1 )
		{
			dd_log(LOG_ERR, "unable to read from source device");
			exit(1);
		}
		
		//
		// change buffer state and signal the other thread (in case it is waiting)
		//
		if ( buffers[i].read_size == 0 )
		{
			dd_log(LOG_INFO,"th_reader buffers[%d] state: done",i);
			buffers[i].state = BUFFER_STATE_DONE;
			is_done = 1;
		}
		else
		{
			dd_log(LOG_DEBUG,"th_reader buffers[%d] state: ddless",i);
			buffers[i].state = BUFFER_STATE_DDLESS;
		}

		buffers[i].read_blocks++;

		//
		// signal the ddless thread that this buffer is ready
		//
		pthread_cond_signal(&buffers[i].ddless_cv);

		//
		// unlock buffer
		//
		pthread_mutex_unlock(&buffers[i].buffer_mutex);

		//
		// next buffer
		//
		i++;
		i = i % READ_BUFFERS;

		//
		// optional read throttle (of course this is highly artificial, because 
		// the read of READ_BUFFER_SIZE is not throttleed, we are simply pausing
		// to make it look like we are throttling after the read is done).
		//
		if ( parms.max_read_mb_sec > 0 )
		{
			gettimeofday(&buffer_time_end,NULL);

			elapsed_seconds  = buffer_time_end.tv_sec  - buffer_time_start.tv_sec;
			elapsed_useconds = buffer_time_end.tv_usec - buffer_time_start.tv_usec;
			elapsed_mtime = ((elapsed_seconds) * 1000 + elapsed_useconds/1000.0) + 0.5;
			read_mb_sec = (double)((READ_BUFFER_SIZE)/MEGABYTE_FACTOR) / ((double)elapsed_mtime/1000);

			if ( read_mb_sec > parms.max_read_mb_sec )
			{
				sleep_factor_usec += 1000;
			}
			else if( read_mb_sec < parms.max_read_mb_sec && sleep_factor_usec > 0)
			{
				sleep_factor_usec -= 1000;
			}
			
			gettimeofday(&buffer_time_start,NULL);
	
			if ( sleep_factor_usec > 0 )
			{
				usleep(sleep_factor_usec);
			}

			dd_log(LOG_DEBUG,"%u MB %ld ms %0.2f MB/s %d usec",
				READ_BUFFER_SIZE / MEGABYTE_FACTOR,
				elapsed_mtime,
				read_mb_sec,
				sleep_factor_usec);
		}
		
	}
	pthread_exit(NULL);
}

//-----------------------------------------------------------------------------
// ddless core thread (calculates checksum and writes)
//-----------------------------------------------------------------------------
void *ddless_thread()
{
	struct checksum_struct *checksum_ptr;
	checksum_ptr = parms.checksum_array;

	u_int64_t	ddless_blocks = 0;

	//
	// enumerate over the buffers
	//
	int is_done = 0;
	int i = 0;
	while(!is_done)
	{
		//
		// lock buffer
		//
		pthread_mutex_lock(&buffers[i].buffer_mutex);
		dd_log(LOG_DEBUG,"th_ddless buffers[%d] lock",i);
		
		//
		// we have a lock and if we find that it is NOT ddless, then we need
		// to wait (the wait does an implicit unlock, so that the other
		// thread can do a lock, process, signal)
		//
		while ( !(buffers[i].state == BUFFER_STATE_DDLESS || 
					buffers[i].state == BUFFER_STATE_DONE) )
		{
			dd_log(LOG_DEBUG,"th_ddless buffers[%d] cond_wait",i);
			pthread_cond_wait(&buffers[i].ddless_cv, &buffers[i].buffer_mutex);
		}
		
		//
		// the reader sets this state on the buffer at which
		// it encounters EOF
		//
		if ( buffers[i].state == BUFFER_STATE_DONE )
		{
			dd_log(LOG_INFO,"th_ddless buffers[%d] state: done found, exiting thread",i);
			is_done = 1;
			break;
		}

		//
		// setup of the aligned buffer pointer
		//
		void *buf;
		buf = buffers[i].aligned_buffer;
		memset(&buffers[i].seg_bytes_dirty_map, 0, sizeof(int) * (BUFFER_SEGMENTS+1));

		//
		// process using SEGMENT_SIZE chunks of data, the last
		// segment might be the full size
		//
		u_int32_t buf_offset = 0;	// buffer offset, incremented usually by SEGMENT_SIZE
		u_int32_t seg_bytes = 0;	// how many bytes we can process per segment
		int segment = 0;
		while ( buf_offset < buffers[i].read_size )
		{
			//
			// handle a read that is less than SEGMENT_SIZE
			//
			seg_bytes = SEGMENT_SIZE;
			if ( buf_offset + SEGMENT_SIZE > buffers[i].read_size )
			{
				seg_bytes = buffers[i].read_size - buf_offset;
				dd_log(LOG_INFO,"processing short segment %d bytes", seg_bytes);
			}

			//
			// compute checksum
			//
			u_int32_t checksum1_murmur;
			u_int32_t checksum2_crc32 = crc32( 0L, Z_NULL, 0 );
			
			checksum1_murmur = MurmurHash2(buf + buf_offset, seg_bytes, 0xbabeaffe);
			checksum2_crc32 = crc32(checksum2_crc32, buf + buf_offset, seg_bytes);
			
			//
			// check if we need to write out this segment of data
			// to the output device (i.e. checksums changed) OR
			// if we have a new checksum file
			//
			if ( parms.checksum_file_new || 
					checksum_ptr->checksum1_murmur != checksum1_murmur ||
					checksum_ptr->checksum2_crc32 != checksum2_crc32)
			{
				if ( parms.runmode == RUNMODE_SOURCE_TARGET )
				{
					//
					// instead of fseek, write the data, we just track the number of
					// bytes to be written within the current segment
					//
					buffers[i].seg_bytes_dirty_map[segment] = seg_bytes;
				}
				
				checksum_ptr->checksum1_murmur = checksum1_murmur;
				checksum_ptr->checksum2_crc32 = checksum2_crc32;
				buffers[i].changed_segments++;
				buffers[i].written_bytes += seg_bytes;
			}
			dd_log(LOG_DEBUG,"checksum_ptr=%p segment=%lu murmur=%08x crc32=%08x",
				checksum_ptr,buf_offset,checksum1_murmur,checksum2_crc32);
			checksum_ptr++;
			buf_offset += SEGMENT_SIZE;
			segment++;
		}

		//
		// write out segments of data, optimising the number of bytes written at once
		//
		int seg_i;
		int active_segment_bytes = 0;
		void *buf_read = NULL;
		u_int64_t write_offset = 0;
		for(seg_i = 0; seg_i < BUFFER_SEGMENTS+1; seg_i++)
		{
			dd_log(LOG_DEBUG,"dirty segment bytes[%d]=%d",
				seg_i,buffers[i].seg_bytes_dirty_map[seg_i]);
			
			//
			// once we encounter non-zero entry we track it and following ones
			//
			if ( buffers[i].seg_bytes_dirty_map[seg_i] != 0 && !active_segment_bytes )
			{
				active_segment_bytes = buffers[i].seg_bytes_dirty_map[seg_i];
				buf_read = buf + (seg_i * SEGMENT_SIZE);
				write_offset = (ddless_blocks * READ_BUFFER_SIZE) + (seg_i * SEGMENT_SIZE);
			}
			//
			// once we encounter a non-zero entry and we are tracking, add the bytes to the
			// active segment byte counter
			//
			else if ( buffers[i].seg_bytes_dirty_map[seg_i] != 0 && active_segment_bytes)
			{
				active_segment_bytes += buffers[i].seg_bytes_dirty_map[seg_i];
			}
			//
			// once we enounter a 0 and are actively tracking, we write the data
			//
			else if ( buffers[i].seg_bytes_dirty_map[seg_i] == 0 && active_segment_bytes)
			{
				dd_log(LOG_DEBUG,"ddless_blocks %llu:  write_offset %llu, write %u bytes", 
					ddless_blocks, write_offset, active_segment_bytes);

				//
				// check if max writes are going to occur, invoke custom command
				//
				if ( parms.max_enabled )
				{
					parms.max_written_bytes += active_segment_bytes;
					if ( parms.max_written_bytes >= parms.max_write_bytes )
					{
						dd_log(LOG_INFO,"next max_written_bytes %llu >= max_write_bytes %llu",
							parms.max_written_bytes, parms.max_write_bytes);
						dd_log(LOG_INFO,"invoking custom command as the next write will be beyond the max write");
						system(parms.max_reached_cmd);
						dd_log(LOG_INFO,"custom command complete, sleeping for 10 seconds");
						sleep(10);
						dd_log(LOG_INFO,"continuing...");

						parms.max_enabled = 0;
						parms.max_reached = 1;
					}
				}

				//
				// position the target file pointer to write_offset
				//
				if ( lseek64(parms.target_fd, write_offset, SEEK_SET) == -1 )
				{
					dd_log(LOG_ERR,"seek set to write offset: %llu failed",write_offset);
					exit(1);
				}

				//
				// now do the actual write
				//
				void *buf_ptr = buf_read;
				ssize_t bytes_write = active_segment_bytes;
				ssize_t bytes_written = 0;
				while(1)
				{
					if ((bytes_written = write(parms.target_fd, buf_ptr, bytes_write))==-1)
					{
						dd_log(LOG_ERR,"write failed");
						exit(1);
					}

					dd_log(LOG_DEBUG,"i=%d buf_read=%p active_segment_bytes=%d buf_ptr=%p bytes_write:%d bytes_written:%d more_to_write:%d",
						i, buf_read,active_segment_bytes,buf_ptr,bytes_write,bytes_written,bytes_written-bytes_write);

					if ( bytes_written-bytes_write == 0 )
						break;
					else
						dd_log(LOG_INFO,"encountered a short write");

					buf_ptr += bytes_written;
					bytes_write += bytes_written;
				}

				//
				// reset the state machine
				//
				active_segment_bytes = 0;
			}
		}

		//
		// increment the block counter
		//
		ddless_blocks++;

		//
		// change buffer state and signal the other thread (in case it is waiting)
		//
		dd_log(LOG_DEBUG,"th_ddless buffers[%d] state: free",i);
		buffers[i].state = BUFFER_STATE_FREE;
		pthread_cond_signal(&buffers[i].read_cv);

		//
		// unlock buffer
		//
		pthread_mutex_unlock(&buffers[i].buffer_mutex);

		//
		// next buffer
		//
		i++;
		i = i % READ_BUFFERS;
	}
	pthread_exit(NULL);
}

//-----------------------------------------------------------------------------
// display configuration parameters
//-----------------------------------------------------------------------------
void dd_parms()
{
	printf("AUTHOR=Steffen Plotner\n");
	printf("RELEASE_DATE=%s\n",RELEASE_DATE);
	printf("READ_BUFFERS=%d\n",READ_BUFFERS);
	printf("READ_BUFFER_SIZE_BYTES=%d\n",READ_BUFFER_SIZE);
	printf("SEGMENT_SIZE_BYTES=%d\n",SEGMENT_SIZE);
}

//-----------------------------------------------------------------------------
// help
//-----------------------------------------------------------------------------
void usage()
{
	printf(
"ddless by Steffen Plotner, release date: "RELEASE_DATE"\n"
"\n"
"Copy source to target keeping track of the segment checksums. Subsequent\n"
"copies are faster because we assume that not all of the source blocks change.\n"
"\n"
"	ddless	[-d] -s <source> [-r <read_rate_mb_s>] -c <checksum> [-b] -t <target>\n"
"		[-m <max_change_gigabytes> -i <cmd>] [-v]\n"
"\n"
"Produce a checksum file using the specified device. Hint: the device could be\n"
"source or target. Use the target and a new checksum file, then compare it to\n"
"the existing checksum file to ensure data integrity of the target.\n"
"\n"
"	ddless	[-d] -s <source> -c <checksum> [-v]\n"
"\n"
"Determine disk read speed zones, outputs data to stdout.\n"
"\n"
"	ddless	[-d] -s <source> [-v]\n"
"\n"
"Outputs the built in parameters\n"
"\n"
"	ddless	-p\n"
"\n"
"Parameters\n"
"	-d	direct io enabled (i.e. bypasses buffer cache)\n"
"\n"
"	-s	source device\n"
"	-r	read rate of source device in megabytes/sec\n"
"	-c	checksum file\n"
"	-b	bail out with exit code 3 because a new checksum file is\n"
"		required, no data is copied from source to target\n"
"	-t	target device\n"
"\n"
"	-m	max number of bytes to be written to target device, when\n"
"		reached invoke the command defined by parameter -i\n"
"	-i	interrupt writing when -m limit is reached and invoke this\n"
"		command to, for example, lvextend snap or lvremove snap\n"
"		of target device\n"
"\n"
"	-p	display parameters (segment size is known as chunksize in LVM2)\n"
"	-v	verbose\n"
"	-vv	verbose+debug\n"
"\n"
"Exit codes:\n"
"	0	successful\n"
"	1	a runtime error code, unable to complete task (detailed perror\n"
"		and logical error message are output via stderr\n"
"	2	max number of gigabytes to be written limit was reached,\n"
"		operation successful\n"
"	3	new checksum file required, only returned if -b is specified\n"
);
}

//-----------------------------------------------------------------------------
// main
//-----------------------------------------------------------------------------
int main(int argc, char *argv[])
{
	int c, errflg;
	extern char *optarg;
	extern int optind, optopt;

	int bail_on_new_checksum = 0;
	parms.o_direct = 0;
	parms.max_write_bytes = 0;
	parms.max_read_mb_sec = 0;
	*parms.max_reached_cmd = '\0';
	errflg = 0;
	while ((c = getopt(argc, argv, "ds:r:c:bt:m:i:hvp")) != -1)
	{
		switch (c)
		{
			case 'd':
				parms.o_direct = 1;
				break;
			case 's':
				strncpy(parms.source_dev, optarg, DEV_NAME_LENGTH);
				break;
			case 'r':
				sscanf(optarg,"%d", &parms.max_read_mb_sec);
				break;
			case 'c':
				strncpy(parms.checksum_file, optarg, DEV_NAME_LENGTH);
				break;
			case 'b':
				bail_on_new_checksum = 1;
				break;
			case 't':
				strncpy(parms.target_dev, optarg, DEV_NAME_LENGTH);
				break;
			case 'm':
				sscanf(optarg,"%llu", &parms.max_write_bytes);
				break;
			case 'i':
				strncpy(parms.max_reached_cmd, optarg, MAX_CMD_LENGTH);
				break;
			case 'v':
				dd_loglevel_inc();
				break;
			case 'p':
				dd_parms();
				exit(0);
			case 'h':
			case '?':
				errflg++;
		}
	}
	if (errflg)
	{
		usage();
		exit(1);
	}

	if ( *parms.source_dev && 
		*parms.checksum_file &&
		*parms.target_dev )
	{
		if ( bail_on_new_checksum )
		{
			dd_log(LOG_INFO,"invoking ddless and verify checksum usability...");
			ddless(RUNMODE_NEW_CHECKSUM);
		}
		else
		{
			dd_log(LOG_INFO,"invoking ddless...");
			ddless(RUNMODE_SOURCE_TARGET);
			if ( parms.max_reached )
				exit(2);
		}
		exit(0);
	}
	else if ( *parms.source_dev && 
		*parms.checksum_file )
	{
		dd_log(LOG_INFO,"invoking ddless checksum generation...");
		ddless(RUNMODE_CHECKSUM_ONLY);
		exit(0);
	}
	else if ( *parms.source_dev )
	{
		dd_log(LOG_INFO,"invoking ddzone...");
		ddless(RUNMODE_DDZONE);
		exit(0);
	}
	else
	{
		usage();
		exit(1);
	}
}

